--Routine to evaluate similarity coefficient between 2 strings
--Author: R. M. Forno - Version 1.0 - 2001/08/26
--Based on an algorithm developed by the author in 1973
--Published in Spanish during 1979

--Given two strings (usually ASCII) s and t, the algorithm determines a
-- similarity coefficient between 0 (minimum) and 1 (maximum).
--It should be noted that the result is non-commutative, that is, it is not
-- the same similarity(a, b) than similarity(b, a).
--Unless there is a reason to the contrary, it is advised to compute
-- the two alternatives and take the greater or the average.
--Some experimentation with these routines will lead to better results.

--As this algorithm is rather slow, it is advised to use previously another one
-- in order to reject rapidly cases of low similarity.
--A simple algorithm to that end is also included here.
--This simple algorith is called simil(sequence s, sequence t)
--The main algorithm is called similarity(sequence s, sequence t)
--Previously, it is necessary to execute algorithm initdif() only once.

--A useful routine to extend upper() to national characters is included

--There are some examples in an attached file

constant MAXDIF = 1000, --Used to avoid computation with floating numbers
	 MAXINT = 1073741823,  --Greatest integer allowed
	 ALPHABET ="abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVXYZ" &
	   "遚ؙ" &
	   "",
	--These constants apply mainly to the English, Spanish, French,
	-- Portuguese, Italian and some other languages.
	--It is possible to modify them to adapt to other languages as well.
	 UPPER = "ABCDEFGHIJKLMNOPQRSTUVWXYZ",
	 LOWER = "abcdefghijklmnopqrstuvwxyz줇",
	 UPPERA = "A",
	 LOWERA = "a",
	 UPPERE = "E",
	 LOWERE = "e",
	 UPPERI = "I",
	 LOWERI = "i",
	 UPPERO = "O",
	 LOWERO = "o",
	 UPPERU = "U",
	 LOWERU = "u",
	 VOWELS = "AEIOUaeiouؙ堂",
	 CONSON = "bcdfghjklmnpqrstvwxyzBCDFGHJKLMNPQRSTVWXYZ",
	 RESULT = UPPER &
	   repeat('A', length(UPPERA)) & repeat('E', length(UPPERE)) &
	   repeat('I', length(UPPERI)) & repeat('O', length(UPPERO)) &
	   repeat('U', length(UPPERU)) &
	   repeat('A', length(UPPERA)) & repeat('E', length(UPPERE)) &
	   repeat('I', length(UPPERI)) & repeat('O', length(UPPERO)) &
	   repeat('U', length(UPPERU)),
	 BASE = LOWER & LOWERA & LOWERE & LOWERI & LOWERO & LOWERU &
	   UPPERA & UPPERE & UPPERI & UPPERO & UPPERU,
	 LOWERTOUPPER = 0.1, --Lower to upper and viceversa coefficient
	 ACCENTFACTOR = 0.1 --Accented - non-accented coefficient
sequence difmat, --Matrix of differences
--You can fill this matrix by specifying its coefficients one by one,
-- but this is a tedious process.
	 shiftvect --Vector for the extended_upper routine

include wildcard.e

function round(object o)
    return floor(o + 0.5)
end function

procedure rel(integer a, integer b, atom c)
--Initializes difference matrix for selected pairs of letters
    integer x, y
    atom d
    d = c * (1 - LOWERTOUPPER)
    x = upper(a) + 1
    y = upper(b) + 1
    a += 1
    b += 1
    difmat[a][b] = d
    difmat[b][a] = d
    difmat[x][y] = d
    difmat[y][x] = d
    difmat[a][y] = c
    difmat[y][a] = c
    difmat[x][b] = c
    difmat[b][x] = c
end procedure

procedure uplow(sequence up, sequence lo, atom c)
--Process upper/lower accented vowels
    atom d
    integer len
    d = c * (1 - LOWERTOUPPER)
    len = length(up)
    for i = 1 to len do
	for j = 1 to len do
	    difmat[up[i] + 1][up[j] + 1] = d
	    difmat[lo[i] + 1][lo[j] + 1] = d
	    difmat[up[i] + 1][lo[j] + 1] = c
	    difmat[lo[i] + 1][up[j] + 1] = c
	end for
    end for
end procedure

global procedure initdif()
--Initializes difference matrix. It should be run only once.
--Same letter cell = 0, very different = 1, others in between.
    sequence s
    integer len
    difmat = repeat(repeat(1, 256), 256)
    rel('b', 'v', 0.6) --Fill cells for selected letter pairs
    rel('c', 'z', 0.6)
    rel('c', 's', 0.6)
    rel('s', 'z', 0.5)
    rel('x', 'c', 0.6)
    rel('x', 's', 0.6)
    rel('x', 'z', 0.6)
    rel('c', 'k', 0.7)
    rel('c', 'q', 0.7)
    rel('k', 'q', 0.6)
    rel('', 'z', 0.7)
    rel('', 'z', 0.7)
    rel('', 's', 0.8)
    rel('', 's', 0.8)
    rel('', 'x', 0.8)
    rel('', 'x', 0.8)
    rel('', 'c', 0.5)
    rel('', 'c', 0.5)
    for i = 1 to length(LOWERI) do --Consider upper() does not work
	rel('y', LOWERI[i], 0.7) -- for national characters 
	rel('', LOWERI[i], 0.7)
	rel('', LOWERI[i], 0.7)
    end for
    for i = 1 to length(LOWERI) do
	rel('j', LOWERI[i], 0.8)
    end for
    for i = 1 to length(LOWERU) do
	rel('w', LOWERU[i], 0.7)
    end for
    rel('w', 'v', 0.7)
    rel('m', 'n', 0.8)
    rel('g', 'j', 0.7)
    rel('d', 't', 0.9)
    rel('h', 'j', 0.95)
    rel('h', 'g', 0.95)
    rel('p', 'b', 0.96)
    rel('p', 'v', 0.97)
    rel('l', 'r', 0.97)
    rel('f', 'v', 0.97)
    rel('f', 'b', 0.98)
    rel('f', 's', 0.98)
    uplow(UPPERA, LOWERA, ACCENTFACTOR)
    uplow(UPPERE, LOWERE, ACCENTFACTOR)
    uplow(UPPERI, LOWERI, ACCENTFACTOR)
    uplow(UPPERO, LOWERO, ACCENTFACTOR)
    uplow(UPPERU, LOWERU, ACCENTFACTOR)
    for i = 1 to length(UPPER) do
	difmat[UPPER[i] + 1][LOWER[i] + 1] = LOWERTOUPPER
	difmat[LOWER[i] + 1][UPPER[i] + 1] = LOWERTOUPPER
    end for
    len = length(VOWELS)
    for i = 1 to len do
	for j = 1 to len do
	    difmat[VOWELS[i] + 1][VOWELS[j] + 1] *= 0.75
	end for
    end for
    len = length(CONSON)
    for i = 1 to len do
	for j = 1 to len do
	    difmat[CONSON[i] + 1][CONSON[j] + 1] *= 0.7
	end for
    end for
    s = repeat(0, 256)
    for i = 1 to length(ALPHABET) do --Flag alpha chars
	s[ALPHABET[i] + 1] = 1
    end for
    for i = 1 to 256 do --Non-alpha chars nearly equal among them
	for j = 1 to 256 do
	    if s[i] = 0 and s[j] = 0 then
		difmat[i][j] = 0.05
	    end if
	end for
    end for
    for i = 1 to 256 do --Same char has 0 difference
	difmat[i][i] = 0
    end for
    difmat = round(difmat * MAXDIF)
	--Generate vector for the extended_upper routine
    shiftvect = repeat(0, 256)
    for i = 1 to 256 do
	shiftvect[i] = i - 1
    end for
    for i = 1 to length(BASE) do
	shiftvect[BASE[i] + 1] = RESULT[i]
    end for
end procedure

function func(sequence s) --Computes factor for distance between characters
--You may modify it as desired
    s *= s
    return floor(s / (1 + s) * MAXDIF)
end function

global function similarity(sequence s, sequence t) --Main algorithm
--This algorithm is somewhat difficult to explain.
--It is based on the technique of dynamic programming (Bellman / Dreyfus).
--It obtains successive minimums from differences between characters and
-- the distance between them.
    integer lens, lent, a
    sequence bmat, wvec, rmat, aux, aux1
    lens = length(s)
    lent = length(t)
    if lens = 0 or lent = 0 then
	return 0
    end if
    bmat = repeat(repeat(0, lent), lens)
    for i = 1 to lens do
	aux = difmat[s[i] + 1]
	for j = 1 to lent do
	    bmat[i][j] = aux[t[j] + 1]
	end for
    end for
    aux = repeat(0, lent)
    for i = 1 to lent do
	aux[i] = lent - i
    end for
    bmat[lens] += func(aux)
    for i = 1 to lent do
	aux[i] = i - 1
    end for
    wvec = bmat[1] + func(aux)
    aux = repeat(repeat(0, lent), lent)
    for i = 1 to lent do
	for j = 1 to lent do
	    aux[i][j] = j - i + 1
	end for
    end for
    rmat = func(aux)
    for i = 2 to lens do
	aux = wvec
	for j = 1 to lent do
	    aux1 = aux + rmat[j]
	    a = MAXINT
	    for k = 1 to lent do
		if a > aux1[k] then
		    a = aux1[k]
		end if
	    end for
	    wvec[j] = a
	end for
	wvec += bmat[i]
    end for
    a = MAXINT
    for i = 1 to lent do
	if a > wvec[i] then
	    a = wvec[i]
	end if
    end for
    return 1.0 - a / MAXDIF / (lens + 1)
end function

function extended_upper(sequence s) --Extends upper() to national characters
    for i = 1 to length(s) do
	s[i] = shiftvect[s[i] + 1]
    end for
    return s
end function

global function simil(sequence s, sequence t) --Screens low similarity data
    sequence x, y, w, z
    integer c, a, lens, lent
    atom r
    lens = length(s)
    lent = length(t)
    x = extended_upper(s)
    y = extended_upper(t)
    w = repeat(0, 256)
    z = w
    for i = 1 to lens do --Count characters
	w[x[i] + 1] += 1
    end for
    for i = 1 to lent do
	z[y[i] + 1] += 1
    end for
    c = 0
    for i = 1 to 256 do
	a = w[i] - z[i]
	if a < 0 then --Use absolute value
	    c -= a
	else
	    c += a
	end if
    end for
    r = 1.0 - c / (lens + lent)
    if r < 0.5 then --Reject low similarity data
	return r
    end if
    return similarity(s, t)
end function

--END OF FILE--

